// MSRaynsford Jan 2023
// Grove Beginner Kit, Arduino Uno

// Creates a Faux seven segment clock that runs on the OLED display
// The single button operation lets you adjust the time
// There is no battery backup or RTC so the clock relies on the accuracy of the CPU
#include <Arduino.h>
#include <Wire.h>
#include <U8g2lib.h>
#include <TimeLib.h>
#include <FastLED.h>

// Several structures and arrays of values are defined in the header
#include "Structures.h"

// Declare the input pin for the button 
const uint8_t btnPin = 6;

// Define the possible modes of operation and a variable to keep track of them
enum clockModes {Normal, AdjMin, AdjHour}; 
clockModes clockMode = clockModes::Normal;

// Declare the OLED screen on the high speed I2C
U8G2_SSD1306_128X64_NONAME_F_HW_I2C u8g2(U8G2_R0, SCL, SDA, U8X8_PIN_NONE); 


// Occurs once at the start of the program
void setup(void) {
  // Initialise the time to 12 noon, the date is irrelevant
  setTime(12,0,0,1,1,2023);

  // Configure the button pin to be a pulled up input
  pinMode(btnPin, INPUT_PULLUP);

  // Start the OLED screen controller
  u8g2.begin();
}

// Continuously loops while the program is running
void loop(void) {
  // Handy function from the FastLED library that gets called every 50ms
  EVERY_N_MILLISECONDS (50) {

    // Monitor the input state from the single button
    doButton();

    // Draw the appropriate thing to the display
    UpdateScreen();
  }
}

// Checks the current state of the input button to decide if the button has
// been pressed once or if it is being held down
// Also handles debounce of the switch because it is called every 50ms
void doButton() {
  // Declare some static variables to remember the states between loops
  static uint8_t oldState = 0;
  static uint8_t holdCount = 0;

  // Read the current state of the button
  uint8_t newState = digitalRead(btnPin);

  // Check the state of the button last time to decide what is currently happening
  // If the button wasn't pushed last time but is now the button has just been pushed down
  // If the button was pushed last time and is still pushed then it is being held down
  // If the button was pushed but isn't now then the button has been released

  // Start by checking what happened last time
  if(oldState == LOW) {
    if(newState == HIGH)
      // The button has just been pushed down
      holdCount = 0;
  }
  else {  
    if(newState == LOW) {
      // The button has just been released
      // Confirm that the button hasn't been used for a 'held' event
      if(holdCount < 10)
        // Adjust the value on the display based on the current mode
        AdjustValue();
    }
    else {
      // The button is being held down
      if(holdCount < 10) {
        // Count how long the button has been held down for
        holdCount += 1;

        // If the button is held down for 10 loops (half a second) then trigger an event
        if(holdCount == 10)
          // Change the mode the clock is running in
          ChangeMode();
      }
    }
  }
  
  // Store the state of the button ready for the next loop
  oldState = newState;
}

// Looks at the current mode of the clock and decides what the next mode should be
// Could be replaced with a single Mode+1 check?
void ChangeMode() {
  switch (clockMode){
    case clockModes::Normal:      clockMode = clockModes::AdjMin;      break;
    case clockModes::AdjMin:      clockMode = clockModes::AdjHour;     break;
    case clockModes::AdjHour:     clockMode = clockModes::Normal;      break;
  }
}

// Updates the image on the screen depending on the current mode of the clock
void UpdateScreen() {
  // Some handy variable for keeping track of the system state
  static uint8_t blinkCount = 0;
  static bool blink = true;

  // Update the count to record how many times this loop has been called
  blinkCount += 1;
  if(blinkCount > 10) {
     blinkCount = 0;

    // Update the blink variable every half a second (10x50ms)
    blink = !blink;
  }

  // Start by clearing the whole display ready for the new image
  u8g2.clearBuffer();  

  // In AdjMin operation the display will blink the value for the mins to indicate that's the value being changed
  // The value will blink for 100ms every half a second
  // Don't draw the minute value in the first two loops
  if(!((clockMode == clockModes::AdjMin) && (blinkCount < 2))) {
    // Draw the minutes with a leading zero
    drawDigit(digitPos[0], minute()); 
    drawDigit(digitPos[1], minute()/10);
  }

  // In AdjMin operation the display will blink the value for the hours to indicate that's the value being changed
  // The value will blink for 100ms every half a second
  // Don't draw the hour value in the first two loops
  if(!((clockMode == clockModes::AdjHour) && (blinkCount < 2))) {
    // Draw the hours with a leading zero
    drawDigit(digitPos[2], hour()); 
    drawDigit(digitPos[3], hour()/10);
  }

  // Draw the middle dash still blinking
  if(blink) drawSegment(65, 35, true);

  // Send the new image to the display
  u8g2.sendBuffer();
}

// Update the Hour or Minute value depending on the current mode of the clock
// The current time has to be reconstructed from parts because the clever
// timeLib will roll over the hour value when you add 1 minute to 59 minutes
void AdjustValue() {
  // Get the current time
  time_t t = now();
  
  // A variable to hold the new value
  uint8_t adj;

  switch (clockMode){    
    case clockModes::AdjMin:      
      // Adjust the time by a minute and roll back to zero if required
      adj = minute(t) + 1;
      if(adj > 59) 
        adj = 0;

      // Rebuild the time
      setTime(hour(t), adj, second(t), day(t), month(t), year(t));
      break;

    case clockModes::AdjHour:     
      // Adjust the time by an hour and roll back to zero if required
      adj = hour(t) + 1;
      if(adj > 23) 
        adj = 0;

      // Rebuild the time
      setTime(adj, minute(t), second(t), day(t), month(t), year(t));
      break;
  }
}


// Convert a digit to the corresponding lit segments on the display
void drawDigit(point pos, uint8_t digit) {
  // Perform a simple check to ensure that the value of the digit isn't too high for the array
  while(digit>9) 
    digit -= 10;
  
  draw7Seg(pos, litSegments[digit]);
}

// Works out which of the seven segments need to be drawn
void draw7Seg(point pos, uint8_t lit) {
  for(uint8_t i=0; i<7; i++) {
    if(lit>>i & 1) {
      drawSegment(pos.x+segments[i].x, pos.y+segments[i].y, segments[i].vertical);
    }
  }
}

// Draws a segment on the display at the desired coordinates
void drawSegment(uint8_t posX, uint8_t posY, bool vert) {
  // Draws 5 lines on the display, centred around the coordinates 
  // The middle line is the longest, getting shorter further away from the middle 
  for(int8_t i=0; i<5; i+=1){
    if(vert)
      u8g2.drawVLine(posX+i-2, posY-segLengths[i].start, segLengths[i].length);
    else
      u8g2.drawHLine(posX-segLengths[i].start, posY+i-2, segLengths[i].length);
  }
}
